__author__ = "Cosmas Heiss, Philipp Flotho"

import moderngl
import numpy as np
from scipy.interpolate import griddata
import cv2


def warp_image_pc(img, flow):
    m, n = img.shape[:2]
    xi, yi = np.meshgrid(np.arange(n).astype(np.float), np.arange(m).astype(np.float))

    tmp = np.empty(img.shape, img.dtype)
    for i in range(3):
        griddata_result = griddata((xi.flatten() + flow[:, :, 0].flatten(),
                                    yi.flatten() + flow[:, :, 1].flatten()),
                                   img[:, :, i].flatten(), (xi.flatten(), yi.flatten()),
                                   method="linear", fill_value=0)
        tmp[:, :, i] = np.reshape(griddata_result, (m, n))

    return tmp


def warp_image_pc_single(img, flow):
    m, n = img.shape[:2]
    xi, yi = np.meshgrid(np.arange(n).astype(np.float), np.arange(m).astype(np.float))

    tmp = np.empty(img.shape, img.dtype)

    griddata_result = griddata((xi.flatten() + flow[:, :, 0].flatten(),
                                yi.flatten() + flow[:, :, 1].flatten()),
                               img[:, :].flatten(), (xi.flatten(), yi.flatten()),
                               method="nearest", fill_value=0)
    tmp[:, :] = np.reshape(griddata_result, (m, n))

    return tmp


def warp_image_backwards(img, flow):
    h, w = flow.shape[:2]
    # flow = -flow
    tmp_flow = np.empty(flow.shape, np.array(flow).dtype)
    np.copyto(tmp_flow, flow)
    tmp_flow[:, :, 0] += np.arange(w)
    tmp_flow[:, :, 1] += np.arange(h)[:, np.newaxis]
    res = cv2.remap(img, tmp_flow, None, cv2.INTER_LINEAR)
    return res


# moderngl-based fast forward warping, author: Cosmas Heiß
class OnlineFrameWarper:
    def __init__(self, image_size):
        self.image_size = image_size
        self.strip_indices = self.generate_triangle_strip_index_array()
        self.ctx = moderngl.create_context(standalone=True, require=330)
        self.ctx.enable(moderngl.DEPTH_TEST)

        self.prog = self.ctx.program(
            vertex_shader="""
                #version 330

                in vec2 in_vert;
                in vec3 in_color;
                in float in_depth;
                out vec3 v_color;


                void main() {
                    v_color = in_color;
                    gl_Position = vec4(in_vert, in_depth, 1.0);
                }
            """,
            fragment_shader="""
                #version 330

                in vec3 v_color;

                out vec3 f_color;

                void main() {
                    f_color = v_color;
                }
            """,
        )
        dummy_vertices = self.get_dummy_vertices()
        self.vertex_buffer = self.ctx.buffer(dummy_vertices)
        self.vertex_array = self.ctx.vertex_array(self.prog, self.vertex_buffer, 'in_vert', 'in_color', 'in_depth')
        self.frame_buffer = self.ctx.simple_framebuffer(image_size[::-1])

    def get_dummy_vertices(self):
        image = np.zeros((*self.image_size, 3))
        xx, yy = np.meshgrid(np.linspace(-1, 1, self.image_size[1], endpoint=True), np.linspace(-1, 1, self.image_size[0], endpoint=True))
        displacements = np.stack((xx, yy), axis=2)
        depth = np.zeros(self.image_size)
        return self.get_into_vertex_buffer_shape(image, displacements, depth)

    def vertices_astype(self, vertices):
        return vertices.astype('f4').tobytes()

    def get_into_vertex_buffer_shape(self, image, displacements, depth):
        vertices = np.concatenate((displacements, image, depth[:, :, None]), axis=2)[self.strip_indices[:, 0], self.strip_indices[:, 1]]
        return self.vertices_astype(vertices)
        # return self.get_into_vb_fast(image, displacements, depth)

    #def get_into_vb_fast(self, image, displacements, depth):
    #    image = cp.asarray(image)
    #    displacements = cp.asarray(displacements)
    #    depth = cp.asarray(depth)
    #    strip_indices = cp.asarray(self.strip_indices)
    #    vertices = cp.concatenate((displacements, image, depth[:, :, None]), axis=2)[strip_indices[:, 0], strip_indices[:, 1]]
    #    return self.vertices_astype(cp.asnumpy(vertices))

    def generate_triangle_strip_index_array(self):
        out_indices_x = []
        out_indices_y = []
        for i in range(self.image_size[1]-1):
            is_reversed = int((i % 2) * (-2) + 1)
            out_indices_x.append(np.arange(self.image_size[0]).repeat(2)[::is_reversed])
            out_indices_y.append(np.tile(np.array([i, i+1]), self.image_size[0]))

        out_indices_x = np.concatenate(out_indices_x).astype(int)
        out_indices_y = np.concatenate(out_indices_y).astype(int)
        return np.stack((out_indices_x, out_indices_y), axis=1)

    def read_frame_buffer(self):
        return np.frombuffer(self.frame_buffer.read(), 'uint8').reshape(self.image_size[0], self.image_size[1], -1)

    def pixel_to_screenspace_coords(self, displacements):
        out = np.zeros_like(displacements)
        out[:, :, 1] = displacements[:, :, 0] * (2.0 / (self.image_size[0] - 1)) - 1.0
        out[:, :, 0] = displacements[:, :, 1] * (2.0 / (self.image_size[1] - 1)) - 1.0
        return out

    def warp_image(self, image, displacements, depth):
        assert image.dtype == displacements.dtype == depth.dtype == float
        assert np.all(np.logical_and(image <= 1.0, image >= 0.0))
        assert image.shape[:2] == displacements.shape[:2] == depth.shape == self.image_size

        displacements = self.pixel_to_screenspace_coords(displacements)
        if depth.max() != depth.min():
            depth = - 0.99 * (depth - depth.min()) / (depth.max() - depth.min())

        self.frame_buffer.use()
        self.frame_buffer.clear(0.0, 0.0, 0.0, 0.0)

        self.vertex_buffer.write(self.get_into_vertex_buffer_shape(image, displacements, depth))

        self.vertex_array.render(moderngl.TRIANGLE_STRIP)

        return self.read_frame_buffer()[:, :, :3]

    def warp_image_uv(self, image, uv, depth=None):
        m, n = uv.shape[0:2]

        x, y = np.meshgrid(np.arange(n).astype(float), np.arange(m).astype(float))

        tmp_flow = np.empty(uv.shape, float)
        tmp_flow[:, :, 0] = y + uv[:, :, 1]
        tmp_flow[:, :, 1] = x + uv[:, :, 0]

        if depth is None:
            depth = np.zeros((m, n), float)

        return (self.warp_image(np.array(image).astype(float) / 255.0, tmp_flow, depth))



import torch
import torch.nn as nn
import torch.nn.functional as F


class FlowHead(nn.Module):
    def __init__(self, input_dim=128, hidden_dim=256):
        super(FlowHead, self).__init__()
        self.conv1 = nn.Conv2d(input_dim, hidden_dim, 3, padding=1)
        self.conv2 = nn.Conv2d(hidden_dim, 2, 3, padding=1)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        return self.conv2(self.relu(self.conv1(x)))

class ConvGRU(nn.Module):
    def __init__(self, hidden_dim=128, input_dim=192+128):
        super(ConvGRU, self).__init__()
        self.convz = nn.Conv2d(hidden_dim+input_dim, hidden_dim, 3, padding=1)
        self.convr = nn.Conv2d(hidden_dim+input_dim, hidden_dim, 3, padding=1)
        self.convq = nn.Conv2d(hidden_dim+input_dim, hidden_dim, 3, padding=1)

    def forward(self, h, x):
        hx = torch.cat([h, x], dim=1)

        z = torch.sigmoid(self.convz(hx))
        r = torch.sigmoid(self.convr(hx))
        q = torch.tanh(self.convq(torch.cat([r*h, x], dim=1)))

        h = (1-z) * h + z * q
        return h

class SepConvGRU(nn.Module):
    def __init__(self, hidden_dim=128, input_dim=192+128):
        super(SepConvGRU, self).__init__()
        self.convz1 = nn.Conv2d(hidden_dim+input_dim, hidden_dim, (1,5), padding=(0,2))
        self.convr1 = nn.Conv2d(hidden_dim+input_dim, hidden_dim, (1,5), padding=(0,2))
        self.convq1 = nn.Conv2d(hidden_dim+input_dim, hidden_dim, (1,5), padding=(0,2))

        self.convz2 = nn.Conv2d(hidden_dim+input_dim, hidden_dim, (5,1), padding=(2,0))
        self.convr2 = nn.Conv2d(hidden_dim+input_dim, hidden_dim, (5,1), padding=(2,0))
        self.convq2 = nn.Conv2d(hidden_dim+input_dim, hidden_dim, (5,1), padding=(2,0))


    def forward(self, h, x):
        # horizontal
        hx = torch.cat([h, x], dim=1)
        z = torch.sigmoid(self.convz1(hx))
        r = torch.sigmoid(self.convr1(hx))
        q = torch.tanh(self.convq1(torch.cat([r*h, x], dim=1)))        
        h = (1-z) * h + z * q

        # vertical
        hx = torch.cat([h, x], dim=1)
        z = torch.sigmoid(self.convz2(hx))
        r = torch.sigmoid(self.convr2(hx))
        q = torch.tanh(self.convq2(torch.cat([r*h, x], dim=1)))       
        h = (1-z) * h + z * q

        return h

class SmallMotionEncoder(nn.Module):
    def __init__(self, args):
        super(SmallMotionEncoder, self).__init__()
        cor_planes = args.corr_levels * (2*args.corr_radius + 1)**2
        self.convc1 = nn.Conv2d(cor_planes, 96, 1, padding=0)
        self.convf1 = nn.Conv2d(2, 64, 7, padding=3)
        self.convf2 = nn.Conv2d(64, 32, 3, padding=1)
        self.conv = nn.Conv2d(128, 80, 3, padding=1)

    def forward(self, flow, corr):
        cor = F.relu(self.convc1(corr))
        flo = F.relu(self.convf1(flow))
        flo = F.relu(self.convf2(flo))
        cor_flo = torch.cat([cor, flo], dim=1)
        out = F.relu(self.conv(cor_flo))
        return torch.cat([out, flow], dim=1)

class BasicMotionEncoder(nn.Module):
    def __init__(self, args):
        super(BasicMotionEncoder, self).__init__()
        cor_planes = args.corr_levels * (2*args.corr_radius + 1)**2
        self.convc1 = nn.Conv2d(cor_planes, 256, 1, padding=0)
        self.convc2 = nn.Conv2d(256, 192, 3, padding=1)
        self.convf1 = nn.Conv2d(2, 128, 7, padding=3)
        self.convf2 = nn.Conv2d(128, 64, 3, padding=1)
        self.conv = nn.Conv2d(64+192, 128-2, 3, padding=1)

    def forward(self, flow, corr):
        cor = F.relu(self.convc1(corr))
        cor = F.relu(self.convc2(cor))
        flo = F.relu(self.convf1(flow))
        flo = F.relu(self.convf2(flo))

        cor_flo = torch.cat([cor, flo], dim=1)
        out = F.relu(self.conv(cor_flo))
        return torch.cat([out, flow], dim=1)

class SmallUpdateBlock(nn.Module):
    def __init__(self, args, hidden_dim=96):
        super(SmallUpdateBlock, self).__init__()
        self.encoder = SmallMotionEncoder(args)
        self.gru = ConvGRU(hidden_dim=hidden_dim, input_dim=82+64)
        self.flow_head = FlowHead(hidden_dim, hidden_dim=128)

    def forward(self, net, inp, corr, flow):
        motion_features = self.encoder(flow, corr)
        inp = torch.cat([inp, motion_features], dim=1)
        net = self.gru(net, inp)
        delta_flow = self.flow_head(net)

        return net, None, delta_flow

class BasicUpdateBlock(nn.Module):
    def __init__(self, args, hidden_dim=128, input_dim=128):
        super(BasicUpdateBlock, self).__init__()
        self.args = args
        self.encoder = BasicMotionEncoder(args)
        self.gru = SepConvGRU(hidden_dim=hidden_dim, input_dim=128+hidden_dim)
        self.flow_head = FlowHead(hidden_dim, hidden_dim=256)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64*9, 1, padding=0))

    def forward(self, net, inp, corr, flow, upsample=True):
        motion_features = self.encoder(flow, corr)
        inp = torch.cat([inp, motion_features], dim=1)

        net = self.gru(net, inp)
        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = .25 * self.mask(net)
        return net, mask, delta_flow






import yolov5_face.detect_face as yf
import torch
import cv2
from neurovc.util import normalize_color
import numpy as np
import os
import requests
from pathlib import Path
from tqdm import tqdm


class _ModelDownloader:
    def __init__(self, model_name, save_dir="~/.neurovc/models"):
        self.model_name = model_name
        self.save_dir = Path(os.path.expanduser(save_dir))
        self.file_id = _file_id_map.get(model_name)
        if not self.file_id:
            raise ValueError(f"Model name '{model_name}' is not valid. Check the file_id_map.")
        self.model_url = f"https://drive.google.com/uc?export=download&id={self.file_id}"
        self.model_path = self.save_dir / f"{model_name}.pt"

    def download_model(self):
        self.save_dir.mkdir(parents=True, exist_ok=True)

        if self.model_path.exists():
            return self.model_path

        print(f"Downloading {self.model_name}...")
        response = requests.get(self.model_url, stream=True)
        if response.status_code == 200:
            total_size = int(response.headers.get('content-length', 0))
            with open(self.model_path, "wb") as f, tqdm(
                desc=f"Downloading {self.model_name}",
                total=total_size,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            ) as bar:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)
                    bar.update(len(chunk))
            print(f"Model downloaded to {self.model_path}")
        else:
            raise Exception(f"Failed to download model: {response.status_code}")
        return self.model_path


def _prepare_model(model_name="YOLOv5n-Face"):
    downloader = _ModelDownloader(model_name)
    model_path = downloader.download_model()
    return model_path


class LandmarkWrapper:
    def process(self, img):
        return self.get_landmarks(img), None


class TFWLandmarker(LandmarkWrapper):
    def __init__(self, model_name="YOLOv5n-Face"):
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        model_path = _prepare_model(model_name)
        self.model = yf.load_model(model_path, self.device)

    def detect(self, img):
        img = normalize_color(img, color_map=cv2.COLORMAP_BONE)
        results = yf.detect_landmarks(self.model, img, self.device)
        return results

    def get_landmarks(self, img):
        results = self.detect(img)
        if len(results) == 0:
            return np.full((5, 2), -1)
        lm = results[0]['landmarks']
        lm = np.array(lm).reshape((-1, 2))
        return lm


_file_id_map = {
    "YOLOv5n": "1PLUq7WbOWS7Ve2VKW7_WBkC3Uksje8Fx",
    "YOLOv5n6": "1wV9t5uH_eiy7WaHdQdWnbeEIijuDAdKI",
    "YOLOv5s": "1IdsdR1-qUeRo5EKQJzGQmRDi2SrMXJG5",
    "YOLOv5s6": "1YZX3t7cSPnWWoic7oJo86ljBQgE5PPb2",
    "YOLOv5n-Face": "1vXk9P3CfhUtRBGI44SqWbuiTJ7rAI4hP"
}
